"""
Apple ID Python
Public Key Module
author: hugh@blinkybeach.com
"""
from typing import TypeVar, Type
from urllib.request import Request, urlopen
from typing import List, Dict, Optional
from rsa import PublicKey as RSA_PublicKey
from siwa.library.data import Data
import json
from siwa.library.key_protocol import PublicKey
from siwa.library.key_cache import KeyCache

T = TypeVar('T', bound='ApplePublicKey')


class ApplePublicKey(PublicKey):

    _RETRIEVAL_URL = 'https://appleid.apple.com/auth/keys'

    def __init__(
        self,
        algorithm: str,
        identifier: str,
        family: str,
        use: str,
        modulus: str,     # n value
        exponent: str     # e value
    ) -> None:

        self._family = family
        self._identifier = identifier
        self._use = use
        self._algorithm = algorithm
        self._modulus = modulus
        self._exponent = exponent

        return

    identifier = property(lambda s: s._identifier)
    rsa_public_key = property(lambda s: RSA_PublicKey(
        n=int.from_bytes(Data.decode_b64(s._modulus), 'big'),
        e=int.from_bytes(Data.decode_b64(s._exponent), 'big')
    ))

    @classmethod
    def decode(cls: Type[T], data: Dict[str, str]) -> T:
        return cls(
            algorithm=data['alg'],
            identifier=data['kid'],
            family=data['kty'],
            use=data['use'],
            modulus=data['n'],
            exponent=data['e']
        )

    @classmethod
    def decode_many(
        cls: Type[T],
        data: List[Dict[str, str]]
    ) -> List[PublicKey]:
        return [cls.decode(d) for d in data]

    @classmethod
    def retrieve_all(
        cls: Type[T],
        cache: Optional[KeyCache] = None
    ) -> List[PublicKey]:

        request = Request(
            url=cls._RETRIEVAL_URL,
            method='GET'
        )

        response = urlopen(request).read()
        data = json.loads(response)
        keys = cls.decode_many(data['keys'])
        if cache is not None:
            cache.store_many(keys)
        return keys

    @classmethod
    def retrieve_by_id(
        cls: Type[T],
        identifier: str,
        cache: Optional[KeyCache] = None
    ) -> Optional[PublicKey]:

        if cache is not None:
            cached = cache.retrieve(identifier)
            if cached is not None:
                return cached

        all_keys = cls.retrieve_all()
        for key in all_keys:
            if key.identifier == identifier:
                return key
        return None



"""
Sign In with Apple
Data Module
author: hugh@blinkybeach.com
"""
import base64
from typing import Union


class Data:

    @staticmethod
    def decode_b64(data: bytes) -> bytes:
        padded = Data.pad(data)
        return base64.urlsafe_b64decode(padded)

    @staticmethod
    def pad(data: Union[bytes, str]) -> bytes:
        if isinstance(data, str):
            data = data.encode('ascii')

        padding = len(data) % 4

        if padding > 0:
            data += b'=' * (4 - padding)

        return data



"""
Signin With Apple
Verify Token Signature Test
author: hugh@blinkybeach.com
"""
from siwa import IdentityToken, KeyCache
from siwa.tests.test import Test
from siwa.tests.test_result import Success, TestResult


class VerifyTokenSignature(Test):

    NAME = 'Verify an IdentityToken\'s signature'

    def execute(self) -> TestResult:

        cache = KeyCache()

        token = IdentityToken.parse(self.example_token)
        valid = token.is_validly_signed(cache)
        assert valid is True

        # Mess up the signature slightly
        bad_raw_token = self.example_token[:-3] + 'j' + self.example_token[-3:]

        bad_token = IdentityToken.parse(bad_raw_token)
        valid = bad_token.is_validly_signed(cache)
        assert valid is False

        return Success()



"""
Signin With Apple
Retrieve Public Keys Test
author: hugh@blinkybeach.com
"""
from siwa import ApplePublicKey
from siwa.tests.test import Test
from siwa.tests.test_result import Success, TestResult


class RetrievePublicKeys(Test):

    NAME = 'Retrieve Apple\'s public keys'

    def execute(self) -> TestResult:

        all_keys = ApplePublicKey.retrieve_all()
        assert isinstance(all_keys, list)
        assert len(all_keys) > 1
        assert False not in [isinstance(k, ApplePublicKey) for k in all_keys]

        key_id = all_keys[0].identifier

        specific_key = ApplePublicKey.retrieve_by_id(key_id)
        assert isinstance(specific_key, ApplePublicKey)
        assert specific_key.identifier == key_id

        return Success()



from siwa.tests.cases.retrieve_public_key import (
    RetrievePublicKeys
)
from siwa.tests.cases.parse_token import ParseIdentityToken
from siwa.tests.cases.verify_token_signature import (
    VerifyTokenSignature
)



from siwa.library.public_key import ApplePublicKey
from siwa.library.token.token import IdentityToken
from siwa.library.key_protocol import PublicKey
from siwa.library.key_cache import KeyCache
from siwa.library.token.payload import Payload
from siwa.library.token.real_person import RealPerson



"""
Apple ID Python
Key Protocol Module
author: hugh@blinkybeach.com
"""
from rsa import PublicKey as RSA_PublicKey


class PublicKey:
    """Abstract protocol defining behaviour of implemented public keys"""
    identifier: str = NotImplemented
    rsa_public_key: RSA_PublicKey = NotImplemented



"""
Signin With Apple
Header Module
author: hugh@blinkybeach.com
"""
from typing import TypeVar, Type, Dict, Optional
from siwa.library.public_key import ApplePublicKey
from siwa.library.key_protocol import PublicKey
from siwa.library.key_cache import KeyCache

T = TypeVar('T', bound='Header')


class Header:

    def __init__(
        self,
        identifier: str,
        algorithm: str
    ) -> None:

        self._identifier = identifier
        self._algorithm = algorithm

        return

    def retrieve_public_key(
        self,
        key_cache: Optional[KeyCache] = None
    ) -> PublicKey:
        key = ApplePublicKey.retrieve_by_id(
            identifier=self._identifier,
            cache=key_cache
        )
        if key is None:
            raise RuntimeError('Apple PublicKey not found ' + self._identifier)
        return key

    @classmethod
    def decode(cls: Type[T], data: Dict) -> T:
        return cls(
            identifier=data['kid'],
            algorithm=data['alg']
        )



"""
Signin With Apple
Token Module
author: hugh@blinkybeach.com
"""
import jwt
from jwt.exceptions import PyJWTError
from typing import TypeVar, Type, Any, Dict, Union, Optional
from siwa.library.data import Data
import json
from siwa.library.key_cache import KeyCache
from siwa.library.token.header import Header
from siwa.library.token.payload import Payload

T = TypeVar('T', bound='IdentityToken')


class IdentityToken:

    def __init__(
        self,
        header: Header,
        payload: Payload,
        raw_signed_body: bytes,
        signature: bytes,
        raw_token: bytes
    ) -> None:

        self._header = header
        self._payload = payload
        self._raw_signed_body = raw_signed_body
        self._signature = signature
        self._raw_token = raw_token

        return

    payload = property(lambda s: s._payload)

    def is_validly_signed(self, key_cache: Optional[KeyCache] = None) -> bool:

        apple_public_key = self._header.retrieve_public_key(key_cache)
        rsa_key = apple_public_key.rsa_public_key
        pks = rsa_key.save_pkcs1()

        try:
            result = jwt.decode(
                jwt=self._raw_token,
                key=pks,
                verify=True,
                algorithms=['RS256'],
                audience='blinkybeach.Makara'
            )
        except PyJWTError:
            return False

        return result is not None

    @classmethod
    def decode(cls: Type[T], data: Dict[str, Any]) -> T:
        raise NotImplementedError

    @classmethod
    def parse(cls: Type[T], data: Union[bytes, str]) -> T:

        if isinstance(data, str):
            data = data.encode('utf-8')

        raw_header, raw_payload, raw_signature = data.rsplit(b'.')

        header = Header.decode(json.loads(Data.decode_b64(raw_header)))
        payload = Payload.decode(json.loads(Data.decode_b64(raw_payload)))
        signature = Data.decode_b64(raw_signature)

        return cls(
            header=header,
            payload=payload,
            raw_signed_body=b'.'.join((raw_header, raw_payload)),
            signature=signature,
            raw_token=data
        )



"""
Signin With Apple
Header Module
author: hugh@blinkybeach.com
"""
from typing import TypeVar, Type, Dict, Optional
from siwa.library.token.real_person import RealPerson

T = TypeVar('T', bound='Payload')


class Payload:

    def __init__(
        self,
        issuer: str,
        subject: str,
        audience: str,
        issued_at: int,
        expiration_time: int,
        nonce: Optional[str],
        nonce_supported: bool,
        email: str,
        email_verified: bool,
        is_private_email: Optional[bool],
        real_person: Optional[RealPerson]

    ) -> None:

        self._issuer = issuer
        self._subject = subject
        self._audience = audience
        self._issued_at = issued_at
        self._expiration_time = expiration_time
        self._nonce = nonce
        self._nonce_supported = nonce_supported
        self._email = email
        self._email_verified = email_verified
        self._is_private_email = is_private_email
        self._real_person = real_person

        return

    unique_apple_user_id = property(lambda s: s._subject)
    expires_utc_seconds_since_epoch = property(lambda s: s._expiration_time)
    issued_utc_seconds_since_epoch = property(lambda s: s._issued_at)
    email = property(lambda s: s._email)
    email_is_private = property(lambda s: s._is_private_email)
    real_person = property(lambda s: s._real_person)

    @classmethod
    def decode(cls: Type[T], data: Dict) -> T:
        return cls(
            issuer=data['iss'],
            subject=data['sub'],
            audience=data['aud'],
            issued_at=data['iat'],
            expiration_time=data['exp'],
            nonce=data['nonce'] if 'nonce' in data else None,
            nonce_supported=data['nonce_supported'],
            email=data['email'],
            email_verified=data['email_verified'],
            is_private_email=data['is_private_email'] if 'is_private_email' in
            data else None,
            real_person=RealPerson(data['real_user_status']) if
            'real_user_status' in data else None
        )



"""
Signin With Apple
Parse Identity Token Test
author: hugh@blinkybeach.com
"""
from siwa import IdentityToken
from siwa.tests.test import Test
from siwa.tests.test_result import Success, TestResult


class ParseIdentityToken(Test):

    NAME = 'Parse an identity token'

    def execute(self) -> TestResult:

        IdentityToken.parse(self.example_token)

        return Success()



"""
Signin With Apple
Test Module
Author: hugh@blinkybeach.com
"""
from siwa.tests.test_result import TestResult, Failure
from datetime import datetime
from typing import Optional
from siwa.library.command_line import CommandLine
import traceback


class Test:
    """A test of a single unit of Signin With Apple functionality"""
    NAME: str = NotImplemented

    _REPORT = '[{result}] {name} ({elapsed})'
    _FAILURE_ADDENDUM = '\n       {reason}'

    def __init__(self) -> None:

        name = self.NAME
        if not isinstance(name, str):
            raise TypeError('Supply name as .NAME property')

        self._name = name
        self._failure_reason: Optional[str] = None
        self._did_pass_store: Optional[bool] = None
        self._test_start: Optional[datetime] = None
        self._test_end: Optional[datetime] = None

        cl = CommandLine.load()
        example_filepath = cl.require('--example-jwt-file')
        with open(example_filepath) as efile:
            self._example_token = efile.read()

        return

    name = property(lambda s: s._name)
    did_pass = property(lambda s: s._did_pass())
    did_fail = property(lambda s: s._did_fail())
    is_complete = property(lambda s: s._is_complete())
    time_elapsed = property(lambda s: s._time_elapsed())

    example_token = property(lambda s: s._example_token)

    def run(self) -> None:
        """Run the test"""
        self._test_start = datetime.utcnow()
        try:
            result = self.execute()
        except Exception as error:
            self.fail_test(error=error)
            self._test_end = datetime.utcnow()
            return
        self._test_end = datetime.utcnow()
        assert isinstance(result, TestResult)
        if result.DID_PASS:
            self.pass_test()
        else:
            assert isinstance(result, Failure)
            self.fail_test(
                reason=result.description,
                error=result.exception
            )
        return

    def execute(self) -> TestResult:
        """Perform the test procedure"""
        raise NotImplementedError

    def _did_pass(self) -> bool:
        """Return True if this test did pass"""
        if self._did_pass_store is None:
            return False
        assert isinstance(self._did_pass_store, bool)
        return self._did_pass_store

    def _did_fail(self) -> bool:
        """Return True if this test did fail"""
        if self._did_pass_store is None:
            return False
        assert isinstance(self._did_pass_store, bool)
        return not self._did_pass_store

    def _is_complete(self) -> bool:
        """Return True if this test is complete"""
        if self._did_pass_store is None:
            return False
        assert isinstance(self._did_pass_store, bool)
        return True

    def _time_elapsed(self) -> int:
        """Return the test run time in milliseconds"""
        if not self.is_complete:
            raise RuntimeError('Cannot measure runtime of an unfinished test')
        assert isinstance(self._test_start, datetime)
        assert isinstance(self._test_end, datetime)
        delta = self._test_end - self._test_start
        elapsed = int(delta.microseconds / 1000) + int(delta.seconds * 1000)
        return elapsed

    def pass_test(self) -> None:
        """Mark this test as having passed"""
        if self._did_pass_store is not None:
            raise RuntimeError('Test already passed or failed')
        self._did_pass_store = True
        return

    def fail_test(
        self,
        reason: Optional[str] = None,
        error: Optional[Exception] = None
    ) -> None:
        """Mark this test has having failed"""
        if self._did_pass_store is not None:
            raise RuntimeError('Test already passed or failed')
        self._did_pass_store = False
        if reason is not None and not isinstance(reason, str):
            raise TypeError(
                'Reason must be str or None, not ' + str(type(reason))
            )
        self._failure_reason = reason
        if error is not None and not isinstance(error, Exception):
            raise TypeError(
                'error must be None or Exception, not ' + str(type(error))
            )
        self._failure_error = error
        return

    def report(self) -> str:
        """Return a string describing the outcome of this Test"""
        if not self.is_complete:
            raise RuntimeError('Cannot report on an unfinished test')
        result = 'PASS'
        if self.did_fail:
            result = 'FAIL'
        report = self._REPORT.format(
            result=result,
            name=self.name,
            elapsed=self._format_elapsed()
        )
        if self.did_fail and self._failure_reason is not None:
            report += self._FAILURE_ADDENDUM.format(
                reason=self._failure_reason
            )
        if self.did_fail and self._failure_error is not None:
            report += '\n       ' + '       '.join(traceback.format_exception(
                None,
                self._failure_error,
                self._failure_error.__traceback__
            ))
        return report

    def _format_elapsed(self) -> str:
        elapsed = self.time_elapsed
        if elapsed < 1000:
            return '{:,}'.format(elapsed) + 'ms'
        return '{:,}'.format(round(elapsed / 1000, 2)) + 's'



"""
Signin With Apple Python
PyPI Setup Module
author: hugh@blinkybeach.com
"""
from setuptools import setup, find_packages
from os import path
from codecs import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'readme.md'), encoding='utf-8') as readme_file:
    LONG_DESCRIPTION = readme_file.read()

with open(path.join(here, 'version'), encoding='utf-8') as version_file:
    VERSION = version_file.read()

setup(
    name='siwa',
    version=VERSION,
    description='Sign in with Apple',
    long_description=LONG_DESCRIPTION,
    url='https://github.com/procuret/procuret-python',
    author='Hugh Jeremy',
    author_email='hugh@blinkybeach.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries'

    ],
    keywords='library apple siwa signin',
    packages=find_packages(),
    long_description_content_type="text/markdown",
    python_requires='>=3.6',
    install_requires=[
        'PyJWT',
        'rsa'
    ],
    project_urls={
        'Github Repository': 'https://github.com/hwjeremy/siwa-python',
        'About': 'https://github.com/hwjeremy/siwa-python'
    }
)



"""
Apple ID Python
Key Cache Module
author: hugh@blinkybeach.com
"""
from siwa.library.key_protocol import PublicKey
from typing import Optional, Dict, List


class KeyCache:

    def __init__(self) -> None:
        self._stored_keys: Dict[str, PublicKey] = {}
        return

    def store(self, key: PublicKey) -> None:
        self._stored_keys[key.identifier] = key
        return

    def store_many(self, keys: List[PublicKey]) -> None:
        for k in keys:
            self.store(k)
        return

    def retrieve(self, identifier: str) -> Optional[PublicKey]:
        if identifier in self._stored_keys.keys():
            return self._stored_keys[identifier]
        return None
